"""
    Homonim: Radiometric homogenisation of aerial and satellite imagery
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import concurrent.futures
import logging
import pathlib
from collections import OrderedDict
from multiprocessing import cpu_count

import numpy as np
import pandas as pd
from rasterio.warp import Resampling
from tqdm import tqdm

from homonim import utils
from homonim.enums import ProcCrs
from homonim.raster_pair import RasterPairReader

logger = logging.getLogger(__name__)


class RasterCompare:
    """Class to statistically compare an image against a reference."""
    default_config = dict(threads=cpu_count() - 1)

    def __init__(self, src_filename, ref_filename, proc_crs=ProcCrs.auto, threads=default_config['threads']):
        """
        Construct the RasterCompare object

        Parameters
        ----------
        src_filename: pathlib.Path, str
            Path to the source image file
        ref_filename: pathlib.Path, str
            Path to the reference image file (whose spatial extent should cover that of src_filename)
        proc_crs: homonim.enums.ProcCrs, optional
            The image CRS and resolution in which to perform the comparison.  proc_crs=ProcCrs.auto will
            automatically choose the lowest resolution of the source and reference CRS's (recommended)
        threads: int, optional
            The number of threads to use for concurrent processing of bands (requires more memory).  0 = use all cpus.
        """
        self._src_filename = pathlib.Path(src_filename)
        self._ref_filename = pathlib.Path(ref_filename)
        self._threads = utils.validate_threads(threads)

        # check src and ref image validity via RasterPairReader and get proc_crs
        # self._raster_pair is initialised to read in bands (not blocks)
        if not isinstance(proc_crs, ProcCrs):
            raise ValueError("'proc_crs' must be an instance of homonim.enums.ProcCrs")
        self._raster_pair = RasterPairReader(self._src_filename, self._ref_filename, proc_crs=proc_crs)
        self._proc_crs = self._raster_pair.proc_crs

    """dict specifying stats labels and functions"""
    _stats = [
        dict(ABBREV='r2',
             DESCRIPTION='Pearson\'s correlation coefficient',
             fn=lambda v1, v2: float(np.corrcoef(v1, v2)[0, 1])),
        dict(ABBREV='RMSE',
             DESCRIPTION='Root Mean Square Error',
             fn=lambda v1, v2: float(np.sqrt(np.mean((v1 - v2) ** 2)))),
        dict(ABBREV='rRMSE',
             DESCRIPTION='Relative RMSE (RMSE/mean(ref))',
             fn=lambda v1, v2: float(np.sqrt(np.mean((v1 - v2) ** 2)) / np.mean(v2))),
        dict(ABBREV='N',
             DESCRIPTION='Number of pixels',
             fn=lambda v1, v2: len(v1))
    ]

    @property
    def stats_key(self):
        """Returns a string of abbreviations and corresponding descriptions for the statistics returned by compare()"""
        return pd.DataFrame(self._stats)[['ABBREV', 'DESCRIPTION']].to_string(index=False, justify="right")

    def compare(self):
        """
        Statistically compare source and reference images and return results.

        Returns
        -------
        res_dict: dict[dict]
                  A dictionary representing the results.
        """
        res_dict = OrderedDict()
        bar_format = '{l_bar}{bar}|{n_fmt}/{total_fmt} bands [{elapsed}<{remaining}]'

        with self._raster_pair as raster_pair:

            def process_band(block_pair):
                """Thread-safe function to process a block (that encompasses the full band)"""
                src_ra, ref_ra = raster_pair.read(block_pair)  # read src and ref bands

                # re-project into the lowest resolution (_proc_crs) space
                if self._proc_crs == ProcCrs.ref:
                    src_ra = src_ra.reproject(**ref_ra.proj_profile, resampling=Resampling.average)
                else:
                    ref_ra = ref_ra.reproject(**src_ra.proj_profile, resampling=Resampling.average)

                mask = src_ra.mask & ref_ra.mask  # combined src and ref mask

                # find stats of valid data
                src_vec = src_ra.array[mask]
                ref_vec = ref_ra.array[mask]

                stats_dict = OrderedDict()
                for _stat in self._stats:
                    stats_dict[_stat['ABBREV']] = _stat['fn'](src_vec, ref_vec)

                # form a string describing the band
                description = (raster_pair.ref_im.descriptions[raster_pair.ref_bands[block_pair.band_i] - 1] or
                        raster_pair.src_im.descriptions[raster_pair.src_bands[block_pair.band_i] - 1] or
                        f'Band {block_pair.band_i + 1}')
                return description, stats_dict

            # process bands concurrently
            with concurrent.futures.ThreadPoolExecutor(max_workers=self._threads) as executor:
                futures = [executor.submit(process_band, block_pair) for block_pair in raster_pair.block_pairs()]

                # wait for threads, get results and raise any thread generated exceptions
                for future in tqdm(concurrent.futures.as_completed(futures), bar_format=bar_format, total=len(futures)):
                    band_desc, band_dict = future.result()
                    res_dict[band_desc] = band_dict

        # use a pandas dataframe to find the mean of the statistics over the bands
        res_df = pd.DataFrame.from_dict(res_dict, orient='index')
        res_df.loc['Mean'] = res_df.mean()
        res_df.N = res_df.N.astype('int')

        return res_df.to_dict(orient='index')



"""
    Homonim: Radiometric homogenisation of aerial and satellite imagery
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import logging
import pathlib
import sys
from timeit import default_timer as timer

import click
import pandas as pd
import rasterio as rio
import yaml
from click.core import ParameterSource
from rasterio.warp import SUPPORTED_RESAMPLING

from homonim import utils
from homonim.compare import RasterCompare
from homonim.enums import ProcCrs, Method
from homonim.fuse import RasterFuse
from homonim.kernel_model import KernelModel
from homonim.stats import ParamStats
from homonim.errors import ImageFormatError

logger = logging.getLogger(__name__)


def _update_existing_keys(default_dict, **kwargs):
    """Update values in a dict with args from matching keys in **kwargs"""
    return {k: kwargs.get(k, v) for k, v in default_dict.items()}


def _configure_logging(verbosity):
    """configure python logging level"""
    # adapted from rasterio https://github.com/rasterio/rasterio
    log_level = max(10, 20 - 10 * verbosity)

    # limit logging config to homonim by applying to package logger, rather than root logger
    # pkg_logger level etc are then 'inherited' by logger = getLogger(__name__) in the modules
    pkg_logger = logging.getLogger(__package__)
    formatter = _PlainInfoFormatter()
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(formatter)
    pkg_logger.addHandler(handler)
    pkg_logger.setLevel(log_level)
    logging.captureWarnings(True)


def _threads_cb(ctx, param, value):
    """click callback to validate threads"""
    try:
        threads = utils.validate_threads(value)
    except Exception as ex:
        raise click.BadParameter(str(ex))
    return threads


def _nodata_cb(ctx, param, value):
    """click callback to convert nodata value to None, nan or float"""
    # adapted from rasterio https://github.com/rasterio/rasterio
    if value is None or value.lower() in ["null", "nil", "none", "nada"]:
        return None
    else:
        # check value is a number and can be cast to output dtype
        try:
            value = float(value.lower())
            if not rio.dtypes.can_cast_dtype(value, ctx.params['dtype']):
                raise click.BadParameter(f"{value} cannot be cast to the output image data type {ctx.params['dtype']}",
                                         param=param, param_hint="nodata")
        except (TypeError, ValueError):
            raise click.BadParameter(f"{value} is not a number", param=param, param_hint="nodata")

        return value


def _creation_options_cb(ctx, param, value):
    """
    click callback to validate `--opt KEY1=VAL1 --opt KEY2=VAL2` and collect
    in a dictionary like the one below, which is what the CLI function receives.
    If no value or `None` is received then an empty dictionary is returned.
        {
            'KEY1': 'VAL1',
            'KEY2': 'VAL2'
        }
    Note: `==VAL` breaks this as `str.split('=', 1)` is used.
    """
    # adapted from rasterio https://github.com/rasterio/rasterio
    if not value:
        return {}
    else:
        out = {}
        for pair in value:
            if '=' not in pair:
                raise click.BadParameter(
                    "Invalid syntax for KEY=VAL arg: {}".format(pair))
            else:
                k, v = pair.split('=', 1)
                k = k.lower()
                v = v.lower()
                out[k] = None if v.lower() in ['none', 'null', 'nil', 'nada'] else yaml.safe_load(v)
        return out


def _param_file_cb(ctx, param, value):
    """click callback to validate parameter image file(s)"""
    for filename in value:
        filename = pathlib.Path(filename)
        try:
            ParamStats._validate_image(filename)
        except (FileNotFoundError, ImageFormatError):
            raise click.BadParameter(f'{filename.name} is not a valid parameter image.', param=param)
    return value


class _PlainInfoFormatter(logging.Formatter):
    """logging formatter to format INFO logs without the module name etc prefix"""

    def format(self, record):
        if record.levelno == logging.INFO:
            self._style._fmt = "%(message)s"
        else:
            self._style._fmt = "%(levelname)s:%(name)s: %(message)s"
        return super().format(record)


class _HomonimCommand(click.Command):
    """
    click Command class that formats single newlines in help strings as single newlines.
    """

    def get_help(self, ctx):
        """Format help strings with single newlines as single newlines."""
        # adapted from https://stackoverflow.com/questions/55585564/python-click-formatting-help-text
        orig_wrap_test = click.formatting.wrap_text

        def wrap_text(text, width, **kwargs):
            text = text.replace('\n', '\n\n')
            return orig_wrap_test(text, width, **kwargs).replace('\n\n', '\n')

        click.formatting.wrap_text = wrap_text
        return click.Command.get_help(self, ctx)


class _FuseCommand(_HomonimCommand):
    """
    click Command class that combines config file and context parameters.

    User-supplied CLI values are given priority, followed by the config file values.
    Where neither user supplied CLI, or config file values exist, parameters retain their defaults.
    """

    # adapted from https://stackoverflow.com/questions/46358797/python-click-supply-arguments-and-options-from-a-configuration-file/46391887
    def invoke(self, ctx):
        config_file = ctx.params['conf']
        if config_file is not None:

            # read the config file into a dict
            with open(config_file) as f:
                config_dict = yaml.safe_load(f)

            for conf_key, conf_value in config_dict.items():
                if conf_key not in ctx.params:
                    raise click.BadParameter(f"Unknown config file parameter '{conf_key}'", ctx=ctx, param_hint="conf")
                else:
                    param_src = ctx.get_parameter_source(conf_key)
                    # overwrite default parameters with values from config file
                    if ctx.params[conf_key] is None or param_src == ParameterSource.DEFAULT:
                        ctx.params[conf_key] = conf_value
                        ctx.set_parameter_source(conf_key, ParameterSource.COMMANDLINE)

        # set the default creation_options if no other driver or creation_options have been specified
        # (this can't be done in a callback as it depends on 'driver')
        if (ctx.get_parameter_source('driver') == ParameterSource.DEFAULT and
                ctx.get_parameter_source('creation_options') == ParameterSource.DEFAULT):
            ctx.params['creation_options'] = RasterFuse.default_out_profile['creation_options']

        return click.Command.invoke(self, ctx)


# define click options and arguments common to more than one command
src_file_arg = click.argument("src-file", nargs=-1, metavar="INPUTS...",
                              type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path))
ref_file_arg = click.argument("ref-file", nargs=1, metavar="REFERENCE",
                              type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path))

threads_option = click.option("-t", "--threads", type=click.INT, default=RasterFuse.default_homo_config['threads'],
                              show_default=True, callback=_threads_cb,
                              help=f"Number of image blocks to process concurrently (0 = use all cpus).")
output_option = click.option("-o", "--output",
                             type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True,
                                             path_type=pathlib.Path),
                             help="Write results to a json file.")


# define the click CLI
@click.group()
@click.option(
    '--verbose', '-v',
    count=True,
    help="Increase verbosity.")
@click.option(
    '--quiet', '-q',
    count=True,
    help="Decrease verbosity.")
def cli(verbose, quiet):
    verbosity = verbose - quiet
    _configure_logging(verbosity)


# fuse command
@click.command(cls=_FuseCommand)
# standard options
@src_file_arg
@ref_file_arg
@click.option("-m", "--method", type=click.Choice(Method, case_sensitive=False),
              default=Method.gain_blk_offset,
              help="Homogenisation method.\ngain: Gain-only model. \ngain-blk-offset: Gain-only model applied to "
                   "offset normalised image blocks [default]. \ngain-offset: Full gain and offset model.")
@click.option("-k", "--kernel-shape", type=click.Tuple([click.INT, click.INT]), nargs=2, default=(5, 5),
              show_default=True, metavar='<HEIGHT WIDTH>',
              help="Kernel height and width in pixels of the --proc-crs / -pc image.")
@click.option("-od", "--output-dir", type=click.Path(exists=True, file_okay=False, writable=True),
              help="Directory in which to create homogenised image(s). [default: use source image directory]")
@click.option("-ovw", "--overwrite", "overwrite", is_flag=True, type=bool, default=False, show_default=True,
              help="Overwrite existing output file(s).")
@click.option("-cmp", "--compare", "do_cmp", type=click.BOOL, is_flag=True, default=False,
              help="Statistically compare source and homogenised images with the reference.")
@click.option("-nbo", "--no-build-ovw", "build_ovw", type=click.BOOL, is_flag=True, default=True,
              help="Turn off overview building for the homogenised image(s).")
@click.option("-pc", "--proc-crs", type=click.Choice(ProcCrs, case_sensitive=False),
              default=ProcCrs.auto,
              help="The image CRS in which to perform parameter estimation.\nauto: Estimate in the lowest resolution "
                   "of the source and reference image CRS's [default - recommended].\nsrc: Estimate in the source image CRS."
                   "\nref: Estimate in the reference image CRS.")
@click.option("-c", "--conf", type=click.Path(exists=True, dir_okay=False, readable=True, path_type=pathlib.Path),
              required=False, default=None, show_default=True,
              help="Path to a yaml configuration file specifying the options below.")
# advanced options
@click.option("-pi", "--param-image", type=click.BOOL, is_flag=True,
              default=RasterFuse.default_homo_config['param_image'],
              help=f"Create a debug image, containing model parameters and R\N{SUPERSCRIPT TWO} values for each "
                   "homogenised image.")
@click.option("-mp", "--mask-partial", type=click.BOOL, is_flag=True,
              default=KernelModel.default_config['mask_partial'],
              help=f"Mask biased homogenised pixels produced from partial kernel or source / reference image coverage.")
@threads_option
@click.option("-mbm", "--max-block-mem", type=click.FLOAT,
              default=RasterFuse.default_homo_config['max_block_mem'], show_default=True,
              help="Maximum image block size in megabytes (0 = block size is the image size).")
@click.option("-ds", "--downsampling", type=click.Choice([r.name for r in rio.warp.SUPPORTED_RESAMPLING]),
              default=KernelModel.default_config['downsampling'], show_default=True,
              help="Resampling method for re-projecting from high to low resolution.")
@click.option("-us", "--upsampling", type=click.Choice([r.name for r in rio.warp.SUPPORTED_RESAMPLING]),
              default=KernelModel.default_config['upsampling'], show_default=True,
              help="Resampling method for re-projecting from low to high resolution.")
@click.option("-rit", "--r2-inpaint-thresh", type=click.FloatRange(min=0, max=1),
              default=KernelModel.default_config['r2_inpaint_thresh'], show_default=True, metavar="FLOAT 0-1",
              help="R\N{SUPERSCRIPT TWO} threshold below which to inpaint model parameters from "
                   "surrounding areas (0 = turn off inpainting). For 'gain-offset' method only.")
@click.option("--out-driver", "driver",
              type=click.Choice(list(rio.drivers.raster_driver_extensions().values()), case_sensitive=False),
              default=RasterFuse.default_out_profile['driver'], show_default=True, metavar="TEXT",
              help="Output image format driver.  See the GDAL docs for options.")
@click.option("--out-dtype", "dtype", type=click.Choice(list(rio.dtypes.dtype_fwd.values())[1:8], case_sensitive=False),
              default=RasterFuse.default_out_profile['dtype'], show_default=True, help="Output image data type.")
@click.option("--out-nodata", "nodata", type=click.STRING, callback=_nodata_cb, metavar="[NUMBER|null|nan]",
              default=RasterFuse.default_out_profile['nodata'], show_default=True,
              help="Output image nodata value.")
@click.option('-co', '--out-profile', 'creation_options', metavar='NAME=VALUE', multiple=True,
              default=(), callback=_creation_options_cb,
              help="Driver specific image creation options for the output image(s).  See the GDAL docs for details.")
@click.pass_context
def fuse(ctx, src_file, ref_file, method, kernel_shape, output_dir, overwrite, do_cmp, build_ovw, proc_crs, conf,
         **kwargs):
    """
    Radiometrically homogenise image(s) by fusion with a reference.

    INPUTS      Path(s) to source image(s) to be homogenised.

    REFERENCE   Path to a surface reflectance reference image.

    The *reference* image bounds should contain those of the *source* image(s), and *source* / *reference* bands should
    correspond i.e. reference band 1 corresponds to source band 1, reference band 2 corresponds to source band 2 etc.

    For best results, the reference and source image(s) should be concurrent, co-located, and spectrally similar.

    \b
    Examples:
    ---------

    Homogenise 'source.tif' with 'reference.tif', using the 'gain-blk-offset' method, and a kernel of 5 x 5 pixels.

    \b
        $ homonim fuse -m gain-blk-offset -k 5 5 source.tif reference.tif


    Homogenise files matching 'source*.tif' with 'reference.tif', using the 'gain-offset' method and a kernel of 15 x 15
    pixels. Place homogenised files in the './homog' directory, produce parameter images, and mask
    partially covered pixels in the homogenised images.

    \b
        $ homonim fuse --method gain-offset --kernel-shape 15 15 -od ./homog
          --param-image --mask-partial source*.tif reference.tif

    """

    try:
        kernel_shape = utils.validate_kernel_shape(kernel_shape, method=method)
    except Exception as ex:
        raise click.BadParameter(str(ex))

    # build configuration dictionaries for ImFuse
    config = dict(homo_config=_update_existing_keys(RasterFuse.default_homo_config, **kwargs),
                  model_config=_update_existing_keys(RasterFuse.default_model_config, **kwargs),
                  out_profile=_update_existing_keys(RasterFuse.default_out_profile, **kwargs))
    compare_files = []

    # iterate over and homogenise source file(s)
    try:
        for src_filename in src_file:
            homo_path = pathlib.Path(output_dir) if output_dir is not None else src_filename.parent

            logger.info(f'\nHomogenising {src_filename.name}')
            with RasterFuse(src_filename, ref_file, homo_path, method=method, kernel_shape=kernel_shape,
                            proc_crs=proc_crs, overwrite=overwrite, **config) as raster_fuse:
                start_time = timer()
                raster_fuse.process()
                # build overviews
                if build_ovw:
                    logger.info(f'Building overviews')
                    raster_fuse.build_overviews()

            logger.info(f'Completed in {timer() - start_time:.2f} secs')
            compare_files += (src_filename, raster_fuse.homo_filename)  # build a list of files to pass to compare

        # compare source and homogenised files with reference (invokes compare command with relevant parameters)
        if do_cmp:
            ctx.invoke(compare, src_file=compare_files, ref_file=ref_file, proc_crs=proc_crs)
    except Exception:
        logger.exception("Exception caught during processing")
        raise click.Abort()


cli.add_command(fuse)


# compare command
@click.command(cls=_HomonimCommand)
@src_file_arg
@ref_file_arg
@click.option("-pc", "--proc-crs", type=click.Choice(ProcCrs, case_sensitive=False),
              default=ProcCrs.auto.name, show_default=True,
              help="The image CRS in which to perform processing.\nauto: process in the lowest "
                   "resolution of the source and reference image CRS's (recommended).\nsrc: process in "
                   "the source image CRS.\nref: process in the reference image CRS.")
@output_option
def compare(src_file, ref_file, proc_crs, output):
    """
    Report similarity statistics between image(s) and a reference.

    INPUTS      Path(s) to image(s) to be compared.

    REFERENCE   Path to a surface reflectance reference image.

    Reference image extents should encompass those of the input image(s), and input / reference bands should correspond
    (i.e. reference band 1 corresponds to input band 1, reference band 2 corresponds to input band
    2 etc).

    \b
    Examples:
    ---------

    Compare 'source.tif' and 'homogenised.tif with 'reference.tif'.

    \b
        $ homonim compare source.tif homogenised.tif reference.tif
    """

    try:
        res_dict = {}
        # iterate over source files, comparing with reference
        for src_filename in src_file:
            logger.info(f'\nComparing {src_filename.name}')
            start_time = timer()
            cmp = RasterCompare(src_filename, ref_file, proc_crs=proc_crs)
            res_dict[str(src_filename)] = cmp.compare()
            logger.info(f'Completed in {timer() - start_time:.2f} secs')

        # print a key for the following tables
        logger.info(f'\n\n{cmp.stats_key}')

        # print a results table per source file
        summary_dict = {}
        for src_file, _res_dict in res_dict.items():
            res_df = pd.DataFrame.from_dict(_res_dict, orient='index')
            res_str = res_df.to_string(float_format="{:.2f}".format, index=True, justify="center",
                                       index_names=False)
            logger.info(f'\n\n{src_file}:\n\n{res_str}')
            summary_dict[src_file] = _res_dict['Mean']

        # print a summary results table comparing all source files
        if len(summary_dict) > 1:
            summ_df = pd.DataFrame.from_dict(summary_dict, orient='index')
            summ_df = summ_df.rename(columns=dict(zip(summ_df.columns, ('Mean ' + summ_df.columns))))
            summ_df.insert(0, 'File', [pathlib.Path(fn).name for fn in summ_df.index])
            summ_str = summ_df.to_string(float_format="{:.2f}".format, index=False, justify="center",
                                         index_names=False)
            logger.info(f'\n\nSummary over bands:\n\n{summ_str}')

        if output is not None:
            res_dict['Reference'] = ref_file.stem
            with open(output, 'w') as file:
                json.dump(res_dict, file)

    except Exception:
        logger.exception("Exception caught during processing")
        raise click.Abort()


cli.add_command(compare)


@click.command()
@click.argument("param-file", nargs=-1, metavar="INPUTS...",
                type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path), callback=_param_file_cb)
@output_option
def stats(param_file, output):
    """
    Report parameter image statistics.

    INPUTS      Path(s) to parameter image(s).
    """

    try:
        stats_dict = {}
        meta_dict = {}

        # process parameter file(s), storing results
        for param_filename in param_file:
            logger.info(f'\nProcessing {param_filename.name}')
            param_stats = ParamStats(param_filename)
            stats_dict[str(param_filename)] = param_stats.stats()
            meta_dict[str(param_filename)] = param_stats.metadata

        # iterate over stored result(s) and print
        for param_filename, param_dict in stats_dict.items():
            param_meta = meta_dict[param_filename]

            logger.info(f'\n{pathlib.Path(param_filename).name}:\n')
            logger.info(param_meta)
            # format the statistics as a dataframe to get printable string
            param_df = pd.DataFrame.from_dict(param_dict, orient='index')
            param_str = param_df.to_string(float_format="{:.2f}".format, index=True, justify="center",
                                           index_names=False)
            logger.info(f'Stats:\n{param_str}')

        if output is not None:
            with open(output, 'w') as file:
                json.dump(stats_dict, file, allow_nan=True)

    except Exception:
        logger.exception("Exception caught during processing")
        raise click.Abort()


cli.add_command(stats)

##



"""
    Homonim: Radiometric homogenisation of aerial and satellite imagery
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


class HomonimError(Exception):
    """Root exception class"""


class UnsupportedImageError(HomonimError):
    """Raised when reading a 12bit jpeg compressed image"""


class ImageContentError(HomonimError):
    """Raised when an image has insufficient coverage or bands"""


class BlockSizeError(HomonimError):
    """Raised when the image block size is invalid"""


class ImageProfileError(HomonimError):
    """Raised when an image profile is invalid"""


class ImageFormatError(HomonimError):
    """Raised when an image format is invalid"""


class IoError(HomonimError):
    """Raised when reading or writing from/to a closed dataset"""


class BandCountMismatchWarning(UserWarning):
    """Warn when the reference and source band counts don't match"""


class NodataMaskWarning(UserWarning):
    """Warn when an image has no mask or nodata values"""



"""
    Homonim: Radiometric homogenisation of aerial and satellite imagery
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import concurrent.futures
import logging
import multiprocessing
import pathlib
import threading
from contextlib import ExitStack
from typing import Tuple

import numpy as np
import rasterio as rio
import rasterio.io
from homonim import utils
from homonim.enums import Method, ProcCrs
from homonim.errors import IoError
from homonim.kernel_model import KernelModel, RefSpaceModel, SrcSpaceModel
from homonim.raster_array import RasterArray
from homonim.raster_pair import RasterPairReader, BlockPair
from rasterio.enums import Resampling
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

logger = logging.getLogger(__name__)


class RasterFuse:
    """
    Class for homogenising ('fusing') a source image with a reference image.
    """

    # default configuration dicts
    default_homo_config = dict(param_image=False, threads=multiprocessing.cpu_count(), max_block_mem=100)
    default_out_profile = dict(driver='GTiff', dtype=RasterArray.default_dtype, nodata=RasterArray.default_nodata,
                               creation_options=dict(tiled=True, blockxsize=512, blockysize=512, compress='deflate',
                                                     interleave='band', photometric=None))
    default_model_config = KernelModel.default_config

    def __init__(self, src_filename, ref_filename, homo_path, method, kernel_shape, proc_crs=ProcCrs.auto,
                 overwrite=True, homo_config=None, model_config=None, out_profile=None):
        """
        Create a RasterFuse class.

        Parameters
        ----------
        src_filename: str, pathlib.Path
            Path to the source image file.
        ref_filename: str, pathlib.Path
            Path to the reference image file.  The extents of this image should cover src_filename with at least a 2
            pixel boundary, and it should have at least as many bands as src_filename.  The ordering of the bands
            in src_filename and ref_filename should match.
        homo_path: str, pathlib.Path
            Path to the homogenised file to create, or a directory in which in which to create an automatically named
            file.
        method: homonim.enums.Method
            The radiometric homogenisation method.
        kernel_shape: tuple
            The (height, width) of the kernel in pixels of the proc_crs image (the lowest resolution image, if
            proc_crs=ProcCrs.auto).
        proc_crs: homonim.enums.ProcCrs, optional
            The initial proc_crs setting, specifying which of the source/reference image spaces should be used for
            parameter estimation .  If proc_crs=ProcCrs.auto (recommended), the lowest resolution image space will be
            used. [default: ProcCrs.auto]
        overwrite: bool, optional
            Overwrite the output file(s) if they exist. [default: True]
        homo_config: dict, optional
            General homogenisation configuration dict with items:
                param_image: bool
                    Turn on/off the production of a debug image containing homogenisation parameters and R2 values.
                threads: int
                    The number of blocks process concurrently (requires more memory). 0 = use all cpus.
                max_block_mem: float
                    An upper limit on the image block size in MB.  Useful for limiting the memory used by each block-
                    processing thread. Note that this is not a limit on the total memory consumed by a thread, as a
                    number of working block-sized arrays are created, but is proportional to the total memory consumed
                    by a thread.
        model_config: dict, optional
            Radiometric modelling configuration dict with items:
                downsampling: rasterio.enums.Resampling, str
                    Resampling method for downsampling.
                upsampling: rasterio.enums.Resampling, str
                    Resampling method for upsampling.
                r2_inpaint_thresh: float
                    R2 threshold below which to inpaint model parameters from "surrounding areas.
                    For 'gain-offset' method only.
                mask_partial: bool
                    Mask homogenised pixels produced from partial kernel or image coverage.
        out_profile: dict, optional
            Output image configuration dict with the items:
                driver: str
                    Output format driver.
                dtype: str
                    Output image data type.
                nodata: float
                    Output image nodata value.
                creation_options: dict
                    Driver specific creation options.  See the rasterio documentation for more information.
        """
        if not isinstance(proc_crs, ProcCrs):
            raise ValueError("'proc_crs' must be an instance of homonim.enums.ProcCrs")
        if not isinstance(method, Method):
            raise ValueError("'method' must be an instance of homonim.enums.Method")
        self._method = method

        self._kernel_shape = utils.validate_kernel_shape(kernel_shape, method=method)
        homo_config['threads'] = utils.validate_threads(homo_config['threads'])

        self._src_filename = pathlib.Path(src_filename)
        self._ref_filename = pathlib.Path(ref_filename)
        self._config = homo_config if homo_config else self.default_homo_config
        self._model_config = model_config if model_config else self.default_model_config
        self._out_profile = out_profile if out_profile else self.default_out_profile

        # get the block overlap for kernel_shape
        overlap = utils.overlap_for_kernel(self._kernel_shape)

        # check the reference and source image validity via RasterPairReader, and get the proc_crs attribute
        raster_pair_args = dict(proc_crs=proc_crs, overlap=overlap, max_block_mem=self._config['max_block_mem'])
        self._raster_pair = RasterPairReader(src_filename, ref_filename, **raster_pair_args)
        self._proc_crs = self._raster_pair.proc_crs

        self._init_out_filenames(pathlib.Path(homo_path), overwrite)

        # create the KernelModel according to proc_crs
        if self._proc_crs == ProcCrs.ref:
            self._model = RefSpaceModel(method=method, kernel_shape=kernel_shape,
                                        param_image=self._config['param_image'], **model_config)
        elif self._proc_crs == ProcCrs.src:
            self._model = SrcSpaceModel(method=method, kernel_shape=kernel_shape,
                                        param_image=self._config['param_image'], **model_config)
        else:
            raise ValueError(f'Invalid proc_crs: {proc_crs}, should be resolved to ProcCrs.ref or ProcCrs.src')

        # initialise other variables
        self._write_lock = threading.Lock()
        self._param_lock = threading.Lock()
        self._out_im = None
        self._param_im = None
        self._stack = None

    def _init_out_filenames(self, homo_path: pathlib.Path, overwrite: bool = True):
        """set up the homogenised and parameter image filenames."""
        homo_path = pathlib.Path(homo_path)
        if homo_path.is_dir():
            # create a filename for the homogenised file in the provided directory
            post_fix = utils.create_homo_postfix(self._proc_crs, self._method, self._kernel_shape,
                                                 self._out_profile['driver'])
            self._homo_filename = homo_path.joinpath(self._src_filename.stem + post_fix)
        else:
            self._homo_filename = homo_path

        self._param_filename = utils.create_param_filename(self._homo_filename)

        if not overwrite and self._homo_filename.exists():
            raise FileExistsError(f"Homogenised image file exists and won't be overwritten without the "
                                  f"'overwrite' option: {self._homo_filename}")
        if not overwrite and self._config['param_image'] and self._param_filename.exists():
            raise FileExistsError(f"Parameter image file exists and won't be overwritten without the "
                                  f"'overwrite' option: {self._param_filename}")

    def _combine_with_config(self, in_profile):
        """Update an input rasterio profile with the configuration profile."""

        if in_profile['driver'].lower() != self._out_profile['driver'].lower():
            # copy only non driver specific keys from input profile when the driver is different to the configured val
            copy_keys = ['driver', 'width', 'height', 'count', 'dtype', 'crs', 'transform']
            out_profile = {copy_key: in_profile[copy_key] for copy_key in copy_keys}
        else:
            out_profile = in_profile.copy()  # copy the whole input profile

        def nested_update(self_dict, other_dict):
            """Update self_dict with a flattened version of other_dict"""
            for other_key, other_value in other_dict.items():
                if isinstance(other_value, dict):
                    # flatten the driver specific nested dict into the root dict
                    nested_update(self_dict, other_value)
                else:
                    self_dict[other_key] = other_value
            return self_dict

        # update out_profile with a flattened self._out_profile
        return nested_update(out_profile, self._out_profile)

    def _create_out_profile(self) -> dict:
        """Create an output image rasterio profile from the source image profile and output configuration"""
        out_profile = self._combine_with_config(self._raster_pair.src_im.profile)
        out_profile['count'] = len(self._raster_pair.src_bands)
        return out_profile

    def _create_param_profile(self) -> dict:
        """Create a debug image rasterio profile from the proc_crs image and configuration"""
        if self._raster_pair.proc_crs == ProcCrs.ref:
            init_profile = self._raster_pair.ref_im.profile
        else:
            init_profile = self._raster_pair.src_im.profile
        param_profile = self._combine_with_config(init_profile)
        # force dtype and nodata to defaults
        param_profile.update(dtype=RasterArray.default_dtype, count=len(self._raster_pair.src_bands) * 3,
                             nodata=RasterArray.default_nodata)
        return param_profile

    def _set_metadata(self, im: rasterio.io.DatasetWriter):
        """Helper function to copy the RasterFuse configuration info to an image (GeoTiff) file."""
        meta_dict = dict(HOMO_SRC_FILE=self._src_filename.name, HOMO_REF_FILE=self._ref_filename.name,
                         HOMO_PROC_CRS=self._proc_crs.name, HOMO_METHOD=self._method.name,
                         HOMO_KERNEL_SHAPE=self._kernel_shape, HOMO_CONF=str(self._config),
                         HOMO_MODEL_CONF=str(self._model.config))
        im.update_tags(**meta_dict)

    def _set_homo_metadata(self, im):
        """
        Copy useful metadata to a homogenised image (GeoTIFF) file.

        Parameters
        ----------
        im: rasterio.io.DatasetWriter
            An open rasterio dataset to write the metadata to.
        """
        self._assert_open()
        self._set_metadata(im)
        for bi in range(0, min(im.count, len(self._raster_pair.ref_bands))):
            ref_bi = self._raster_pair.ref_bands[bi]
            ref_meta_dict = self._raster_pair.ref_im.tags(ref_bi)
            homo_meta_dict = {k: v for k, v in ref_meta_dict.items() if k in ['ABBREV', 'ID', 'NAME']}
            im.set_band_description(bi + 1, self._raster_pair.ref_im.descriptions[ref_bi - 1])
            im.update_tags(bi + 1, **homo_meta_dict)

    def _set_param_metadata(self, im):
        """
        Copy useful metadata to a parameter image (GeoTIFF) dataset.

        Parameters
        ----------
        im: rasterio.io.DatasetWriter
            An open rasterio dataset to write the metadata to.
        """
        # Use reference file band descriptions to make debug image band descriptions
        self._assert_open()
        self._set_metadata(im)
        num_src_bands = len(self._raster_pair.src_bands)
        for bi in range(0, num_src_bands):
            ref_bi = self._raster_pair.ref_bands[bi]
            ref_descr = self._raster_pair.ref_im.descriptions[ref_bi - 1] or f'B{ref_bi}'
            ref_meta_dict = self._raster_pair.ref_im.tags(ref_bi)
            for param_i, param_name in zip(range(bi, im.count, num_src_bands), ['GAIN', 'OFFSET', 'R2']):
                im.set_band_description(param_i + 1, f'{ref_descr}_{param_name}')
                param_meta_dict = {k: f'{v.upper()} {param_name}' for k, v in ref_meta_dict.items() if
                                   k in ['ABBREV', 'ID', 'NAME']}
                im.update_tags(param_i + 1, **param_meta_dict)

    def _assert_open(self):
        """Raise an IoError if the source, reference or homogenised image(s) are not open."""
        if self.closed:
            raise IoError(f'The raster file(s) have not been opened.')

    def open(self):
        """Open the source and reference images for reading, and output image(s) for writing."""
        self._raster_pair.open()
        self._out_im = rio.open(self._homo_filename, 'w', **self._create_out_profile())
        self._set_homo_metadata(self._out_im)

        if self._config['param_image']:
            self._param_im = rio.open(self._param_filename, 'w', **self._create_param_profile())
            self._set_param_metadata(self._param_im)

    def close(self):
        """Close all open images."""
        self._out_im.close()
        if self._param_im:
            self._param_im.close()
        self._raster_pair.close()

    def __enter__(self):
        self._stack = ExitStack()
        self._stack.enter_context(rio.Env(GDAL_NUM_THREADS='ALL_CPUs'))
        self._stack.enter_context(logging_redirect_tqdm([logging.getLogger(__package__)]))
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        self._stack.__exit__(exc_type, exc_val, exc_tb)

    @property
    def method(self) -> Method:
        """Homogenisation method."""
        return self._method

    @property
    def kernel_shape(self) -> Tuple[int, int]:
        """Kernel shape."""
        return tuple(self._kernel_shape)

    @property
    def proc_crs(self) -> ProcCrs:
        """The 'processing CRS' i.e. which of the source/reference image spaces is selected for processing."""
        return self._proc_crs

    @property
    def closed(self) -> bool:
        """True when the RasterFuse images are closed, otherwise False."""
        return not self._out_im or self._out_im.closed or not self._raster_pair or self._raster_pair.closed

    @property
    def homo_filename(self) -> pathlib.Path:
        """Path to the homogenised image file."""
        return self._homo_filename

    @property
    def param_filename(self) -> pathlib.Path:
        """Path to the parameter image file."""
        return self._param_filename if self._config['param_image'] else None

    def _build_overviews(self, im):
        """
        Build internal overviews for a rasterio dataset.

        Parameters
        ----------
        im: rasterio.io.DatasetWriter
            An open rasterio dataset to write the metadata to.
        """

        # limit overviews so that the highest level has at least 2**8=256 pixels along the shortest dimension,
        # and so there are no more than 8 levels.
        max_ovw_levels = int(np.min(np.log2(im.shape)))
        num_ovw_levels = np.min([8, max_ovw_levels - 8])
        ovw_levels = [2 ** m for m in range(1, num_ovw_levels + 1)]
        im.build_overviews(ovw_levels, Resampling.average)

    def build_overviews(self):
        """
        Build overviews for the homogenised and parameter image files.
        Should be called after homogenise().
        """
        self._assert_open()
        self._build_overviews(self._out_im)
        if self._config['param_image']:
            self._build_overviews(self._param_im)

    def _process_block(self, block_pair: BlockPair):
        """Thread-safe function to homogenise a source image block"""

        # read source and reference blocks
        src_ra, ref_ra = self._raster_pair.read(block_pair)
        # fit and apply the sliding kernel models
        param_ra = self._model.fit(ref_ra, src_ra)
        out_ra = self._model.apply(src_ra, param_ra)
        # change the output nodata value so that is is masked correctly for out_im
        out_ra.nodata = self._out_im.nodata

        with self._write_lock:  # write the output block
            out_ra.to_rio_dataset(self._out_im, indexes=block_pair.band_i + 1, window=block_pair.src_out_block)

        if self._config['param_image']:
            with self._param_lock:  # write the parameter block
                indexes = np.arange(param_ra.count) * len(self._raster_pair.src_bands) + block_pair.band_i + 1
                param_out_block = block_pair.ref_out_block if self._proc_crs == ProcCrs.ref else block_pair.src_out_block
                param_ra.to_rio_dataset(self._param_im, indexes=indexes, window=param_out_block)

    def process(self):
        """
        Homogenise the source image in blocks.
        """
        self._assert_open()

        # get block pairs and process
        bar_format = '{l_bar}{bar}|{n_fmt}/{total_fmt} blocks [{elapsed}<{remaining}]'  # tqdm progress bar format
        if self._config['threads'] == 1:
            # process blocks consecutively in the main thread (useful for profiling)
            block_pairs = [block_pair for block_pair in self._raster_pair.block_pairs()]
            for block_pair in tqdm(block_pairs, bar_format=bar_format):
                self._process_block(block_pair)
        else:
            # process blocks concurrently
            with concurrent.futures.ThreadPoolExecutor(max_workers=self._config['threads']) as executor:
                # submit blocks for processing
                futures = [executor.submit(self._process_block, block_pair)
                           for block_pair in self._raster_pair.block_pairs()]

                # wait for threads in order of completion, and raise any thread generated exceptions
                for future in tqdm(concurrent.futures.as_completed(futures), bar_format=bar_format,
                                   total=len(futures)):
                    future.result()



"""
    Homonim: Radiometric homogenisation of aerial and satellite imagery
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
import pathlib
import threading
from collections import namedtuple
from itertools import product
from typing import Tuple

import numpy as np
import rasterio
import rasterio as rio
from homonim import errors
from homonim import utils
from homonim.enums import ProcCrs
from homonim.raster_array import RasterArray
from rasterio.enums import MaskFlags
from rasterio.vrt import WarpedVRT
from rasterio.warp import Resampling
from rasterio.windows import Window

logger = logging.getLogger(__name__)

"""Named tuple to contain a set of matching block windows for a source-reference image pair"""
BlockPair = namedtuple('BlockPair',
                       ['band_i',  # band index (0 based)
                        'src_in_block',  # overlapping source window
                        'ref_in_block',  # overlapping reference window
                        'src_out_block',  # non-overlapping source window
                        'ref_out_block',  # non-overlapping reference window
                        'outer'])  # True if src_*_block touches the boundary of the source image


class RasterPairReader:
    """
    A thread-safe class for reading matching blocks from a source and reference image pair.

    Images are divided into (optionally overlapping) blocks to satisfy a maximum block memory limit.  Blocks extents
    are constructed so that re-projections between source and reference do not lose valid data.
    """

    def __init__(self, src_filename, ref_filename, proc_crs=ProcCrs.auto, overlap=(0, 0), max_block_mem=np.inf):
        """
        Construct a RasterPairReader.

        Parameters
        ----------
        src_filename: str, pathlib.Path
            Path to the source image file.
        ref_filename: str, pathlib.Path
            Path to the reference image file.  The extents of this image should cover src_filename with at least a 2
            pixel boundary, and it should have at least as many bands as src_filename.  The ordering of the bands
            in src_filename and ref_filename should match.
        proc_crs: homonim.enums.ProcCrs
            The initial proc_crs setting, if proc_crs=ProcCrs.auto (recommended), the RasterPairReader.proc_crs
            attribute will be set to represent the lowest resolution of the source and reference image.
            [default: ProcCrs.auto]
        overlap: tuple
            Block overlap (rows, columns) in pixels of the proc_crs image. [default: (0,0)]
        max_block_mem: float
            The maximum allowable block size in MB. The image is divided into 2**n blocks with n the smallest number
            where max_block_mem is satisfied.  If max_block_mem=float('inf'), the block shape will be set to the
            encompass full extent of the source image. [default: float('inf')]
        """
        self._src_filename = pathlib.Path(src_filename)
        self._ref_filename = pathlib.Path(ref_filename)
        if not isinstance(proc_crs, ProcCrs):
            raise ValueError("'proc_crs' must be an instance of homonim.enums.ProcCrs")
        self._proc_crs = proc_crs
        self._overlap = np.array(overlap).astype('int')
        self._max_block_mem = max_block_mem

        self._env = None
        self._src_lock = threading.Lock()
        self._ref_lock = threading.Lock()
        self._src_im = None
        self._ref_im = None
        self._init_image_pair()

    @staticmethod
    def _validate_image(im: rasterio.DatasetReader):
        """Validate an open rasterio dataset for use as a source or reference image."""

        try:
            _ = im.read(1, window=im.block_window(1, 0, 0))
        except Exception as ex:
            if 'compress' in im.profile and im.profile['compress'] == 'jpeg':  # assume it is a 12bit JPEG
                raise errors.UnsupportedImageError(
                    f'Could not read image {im.name}.  JPEG compression with NBITS==12 is not supported, '
                    f'you probably need to recompress this image.'
                )
            else:
                raise ex

        # warn if there is no nodata or associated mask
        is_masked = any([MaskFlags.all_valid not in im.mask_flag_enums[bi] for bi in range(im.count)])
        if im.nodata is None and not is_masked:
            logger.warning(f'{im.name} has no mask or nodata value, '
                           f'any invalid pixels should be masked before processing.')

    @staticmethod
    def _validate_image_pair(src_im: rasterio.DatasetReader, ref_im: rasterio.DatasetReader):
        """Validate a pair of rasterio datasets for use as a source-reference image pair."""

        for im in (src_im, ref_im):
            RasterPairReader._validate_image(im)
        # check reference image extents cover the source
        if not utils.covers_bounds(ref_im, src_im):
            raise errors.ImageContentError(f'Reference extent does not cover source image')

        # retrieve non-alpha bands
        src_bands = utils.get_nonalpha_bands(src_im)
        ref_bands = utils.get_nonalpha_bands(ref_im)

        # check reference has enough bands
        if len(src_bands) > len(ref_bands):
            raise errors.ImageContentError(f'Reference ({ref_im.name}) has fewer non-alpha bands than '
                                           f'source ({src_im.name}).')

        # warn if source and reference band counts don't match
        if len(src_bands) != len(ref_bands):
            logger.warning(f'Source and reference image non-alpha band counts don`t match. '
                           f'Using the first {len(src_bands)} non-alpha bands of reference.')

        # warn if the source and reference are not in the same CRS
        if src_im.crs.to_proj4() != ref_im.crs.to_proj4():
            logger.warning(f'Source and reference image pair are not in the same CRS: {src_im.name} and '
                           f'{ref_im.name}')

    @staticmethod
    def _resolve_proc_crs(src_im: rasterio.DatasetReader, ref_im: rasterio.DatasetReader,
                          proc_crs: ProcCrs = ProcCrs.auto):
        """
        Resolve proc_crs from auto to the lowest resolution of the source and reference image pair.  If it is already
        resolved, then warn if it does not correspond to the lowest resolution image of the pair.
        """

        # compare source and reference resolutions
        src_pixel_smaller = np.prod(src_im.res) < np.prod(ref_im.res)
        cmp_str = "smaller" if src_pixel_smaller else "larger"
        if proc_crs == ProcCrs.auto:
            # set proc_crs to the lowest resolution of the source and reference images
            proc_crs = ProcCrs.ref if src_pixel_smaller else ProcCrs.src
            logger.debug(
                f"Source pixel size {np.round(src_im.res, decimals=3)} is {cmp_str} than the reference "
                f"{np.round(ref_im.res, decimals=3)}. Using proc_crs='{proc_crs}'.")
        elif ((proc_crs == ProcCrs.src and src_pixel_smaller) or
              (proc_crs == ProcCrs.ref and not src_pixel_smaller)):
            # warn if the proc_crs value does not correspond to the lowest resolution of the source and
            # reference images
            rec_crs_str = ProcCrs.ref if src_pixel_smaller else ProcCrs.src
            logger.warning(f"proc_crs={rec_crs_str} is recommended when "
                           f"the source pixel size is {cmp_str} than the reference.")
        return proc_crs

    def _auto_block_shape(self, proc_win: Window = None):
        """Find a block shape that satisfies max_block_mem."""

        # adjust max_block_mem to represent the size of a block in the highest resolution image, but scaled to the
        # equivalent in proc_crs.
        src_pix_area = np.product(self._src_im.res)
        ref_pix_area = np.product(self._ref_im.res)
        if self._proc_crs == ProcCrs.ref:
            mem_scale = src_pix_area / ref_pix_area if ref_pix_area > src_pix_area else 1.
        elif self._proc_crs == ProcCrs.src:
            mem_scale = 1. if ref_pix_area > src_pix_area else ref_pix_area / src_pix_area
        else:
            raise ValueError("'proc_crs' has not been resolved - the raster pair must be opened first.")
        max_block_mem = self._max_block_mem * mem_scale if self._max_block_mem > 0 else np.inf

        max_block_mem *= 2 ** 20  # convert MB to bytes
        dtype_size = np.dtype(RasterArray.default_dtype).itemsize  # the size of the RasterArray data type

        if proc_win is None:
            # set the starting block_shape to correspond to the entire band
            block_shape = (self._ref_im.shape if self._proc_crs == ProcCrs.ref else self._src_im.shape).astype('float')
        else:
            # set the starting block_shape to correspond to the entire window
            block_shape = np.array((proc_win.height, proc_win.width)).astype('float')

        # keep halving the block_shape along the longest dimension until it satisfies max_block_mem
        while (np.product(block_shape) * dtype_size) > max_block_mem:
            div_dim = np.argmax(block_shape)
            block_shape[div_dim] /= 2

        if np.any(block_shape < (1, 1)):
            raise errors.BlockSizeError(f"The auto block shape is smaller than a pixel.  Increase 'max_block_mem'.")

        block_shape = np.ceil(block_shape).astype('int')
        # warn if the block shape in the highest res image is less than a typical tile
        if np.any(block_shape / mem_scale < (256, 256)):
            logger.warning(f"The auto block shape is small: {block_shape}.  Increase 'max_block_mem' to improve "
                           f"processing times.")

        return block_shape

    def _init_image_pair(self):
        """Prepare the class for reading."""
        self.open()
        try:
            self._validate_image_pair(self._src_im, self._ref_im)
            # check and resolve proc_crs
            self._proc_crs = RasterPairReader._resolve_proc_crs(self._src_im, self._ref_im, proc_crs=self._proc_crs)
            # get non-alpha band indices for reading
            self._src_bands = utils.get_nonalpha_bands(self._src_im)
            logger.debug(f'{self._src_filename.name} non-alpha bands: {self._src_bands}')
            self._ref_bands = utils.get_nonalpha_bands(self._ref_im)
            logger.debug(f'{self._ref_filename.name} non-alpha bands: {self._ref_bands}')

            # create image windows that allow re-projections between source and reference without loss of data
            self._ref_win = utils.expand_window_to_grid(self._ref_im.window(*self._src_im.bounds))
            self._src_win = utils.expand_window_to_grid(self._src_im.window(*self._ref_im.window_bounds(self._ref_win)))

            # generate the auto block shape for reading
            proc_win = self._ref_win if self._proc_crs == ProcCrs.ref else self._src_win
            self._block_shape = self._auto_block_shape(proc_win=proc_win)
            logger.debug(f'Auto block shape: {self._block_shape}, of image shape: {[proc_win.height, proc_win.width]}'
                         f' ({self._proc_crs.name} pixels)')
            logger.debug(f'Block overlap: {self._overlap} ({self._proc_crs.name} pixels)')
        finally:
            self.close()

    def _assert_open(self):
        """Raise an IoError if the source and reference images are not open."""
        if self.closed:
            raise errors.IoError(f'The raster pair has not been opened: {self._src_filename.name} and '
                                 f'{self._ref_filename.name}')

    def open(self):
        """Open the source and reference images for reading."""
        self._src_im = rio.open(self._src_filename, 'r')
        self._ref_im = rio.open(self._ref_filename, 'r')

        # It is essential that the source and reference are in the same CRS so that rectangular regions of valid
        # data in one will re-project to rectangular regions of valid data in the other.
        if self._src_im.crs.to_proj4() != self._ref_im.crs.to_proj4():
            # open the image pair in the same CRS, re-projecting the proc_crs (usually lower resolution) image into the
            # CRS of the other
            if self._proc_crs == ProcCrs.ref:
                self._ref_im = WarpedVRT(self._ref_im, crs=self._src_im.crs, resampling=Resampling.bilinear)
            else:
                self._src_im = WarpedVRT(self._src_im, crs=self._ref_im.crs, resampling=Resampling.bilinear)

    def close(self):
        """Close the source and reference image datasets."""
        self._src_im.close()
        self._ref_im.close()

    def __enter__(self):
        self._env = rio.Env(GDAL_NUM_THREADS='ALL_CPUs').__enter__()
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        self._env.__exit__(exc_type, exc_val, exc_tb)

    @property
    def src_im(self) -> rasterio.DatasetReader:
        """The source rasterio dataset."""
        self._assert_open()
        return self._src_im

    @property
    def ref_im(self) -> rasterio.DatasetReader:
        """The reference rasterio dataset."""
        self._assert_open()
        return self._ref_im

    @property
    def src_bands(self) -> Tuple[int,]:
        """The source non-alpha band indices (1-based)."""
        return self._src_bands

    @property
    def ref_bands(self) -> Tuple[int,]:
        """The reference non-alpha band indices (1-based)."""
        return self._ref_bands

    @property
    def proc_crs(self) -> ProcCrs:
        """The 'processing CRS' i.e. which of the source/reference image spaces is selected for processing."""
        return self._proc_crs

    @property
    def overlap(self) -> Tuple[int, int]:
        """The block overlap (rows, columns) in pixels of the 'proc_crs' image."""
        return tuple(self._overlap)

    @property
    def block_shape(self) -> Tuple[int, int]:
        """The image block shape (rows, columns) in pixels of the 'proc_crs' image."""
        return self._block_shape

    @property
    def closed(self) -> bool:
        """True when the RasterPair is closed, otherwise False."""
        return not self._src_im or not self._ref_im or self._src_im.closed or self._ref_im.closed

    def read(self, block):
        """
        Read a matching pair of source and reference image blocks.

        Parameters
        ----------
        block: BlockPair
            The BlockPair named tuple (as returned by block_pairs()) specifying the source and reference blocks to be
            read.

        Returns
        -------
        src_ra: RasterArray
            The source image block wrapped in a RasterArray.
        ref_ra: RasterArray
            The reference image block wrapped in a RasterArray.
        """
        self._assert_open()

        with self._src_lock:
            src_ra = RasterArray.from_rio_dataset(self._src_im, indexes=self._src_bands[block.band_i],
                                                  window=block.src_in_block)
        with self._ref_lock:
            ref_ra = RasterArray.from_rio_dataset(self._ref_im, indexes=self._ref_bands[block.band_i],
                                                  window=block.ref_in_block)
        return src_ra, ref_ra

    def block_pairs(self):
        """
        Iterator over the paired source-reference image blocks.

        Yields
        -------
        block_pair: BlockPair
            A BlockPair named tuple specifying the overlapping ('*_in_block'), and non-overlapping ('*_out_block')
            source and reference image blocks.
        """
        self._assert_open()

        # initialise block formation variables
        # blocks are first formed in proc_crs, then transformed to the 'other' image crs, so here we assign the src/ref
        # windows etc to proc_* equivalents
        if self._proc_crs == ProcCrs.ref:
            proc_win, proc_im, other_im = (self._ref_win, self._ref_im, self._src_im)
        else:
            proc_win, proc_im, other_im = (self._src_win, self._src_im, self._ref_im)
        overlap = self._overlap
        block_shape = self._block_shape
        proc_win_ul = np.array((proc_win.row_off, proc_win.col_off))
        proc_win_br = np.array((proc_win.height + proc_win.row_off, proc_win.width + proc_win.col_off))

        # Create the overlapping blocks in proc_crs, and find their equivalents in 'other' space.
        # Outer loop over bands so that all blocks in a band are yielded consecutively - this is fastest for
        # reading band interleaved images.
        for band_i in range(len(self._src_bands)):
            # Inner loop over the upper left corner row, col for each overlapping block
            ul_row_range = range(proc_win.row_off - overlap[0],
                                 proc_win.row_off + proc_win.height - overlap[0],
                                 block_shape[0])
            ul_col_range = range(proc_win.col_off - overlap[1],
                                 proc_win.col_off + proc_win.width - overlap[1],
                                 block_shape[1])
            for ul_row, ul_col in product(ul_row_range, ul_col_range):
                # find UL and BR corners for overlapping block in proc space
                ul = np.array((ul_row, ul_col))
                br = ul + block_shape + (2 * overlap)
                # limit block extents to image window extents
                in_ul = np.fmax(ul, proc_win_ul)
                in_br = np.fmin(br, proc_win_br)
                # find UL and BR corners for non-overlapping block in proc space
                out_ul = np.fmax(ul + overlap, proc_win_ul)
                out_br = np.fmin(br - overlap, proc_win_br)
                # block touches image boundary?
                outer = np.any(in_ul <= proc_win_ul) or np.any(in_br >= proc_win_br)

                # create rasterio windows corresponding to above block corners
                proc_in_block = Window(*in_ul[::-1], *np.subtract(in_br, in_ul)[::-1])
                proc_out_block = Window(*out_ul[::-1], *np.subtract(out_br, out_ul)[::-1])

                # create equivalent rasterio windows in 'other' space
                other_in_block = utils.expand_window_to_grid(other_im.window(*proc_im.window_bounds(proc_in_block)))
                other_out_block = utils.round_window_to_grid(other_im.window(*proc_im.window_bounds(proc_out_block)))

                # create the BlockPair named tuple, assigning 'proc' and 'other' back to 'src' and 'ref' for passing to
                # read()
                if self._proc_crs == ProcCrs.ref:
                    block_pair = BlockPair(band_i, other_in_block, proc_in_block, other_out_block, proc_out_block,
                                           outer)
                else:
                    block_pair = BlockPair(band_i, proc_in_block, other_in_block, proc_out_block, other_out_block,
                                           outer)
                yield block_pair



"""
    Homonim: Radiometric homogenisation of aerial and satellite imagery
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import pathlib
from collections import OrderedDict
from multiprocessing import cpu_count

import numpy as np
import pandas as pd
import rasterio as rio
from homonim.enums import Method
from rasterio.enums import Resampling, ColorInterp
from rasterio.vrt import WarpedVRT
from rasterio.windows import Window, get_data_window

logger = logging.getLogger(__name__)


def nan_equals(a, b):
    """Compare two numpy objects a & b, returning true where elements of both a & b are nan"""
    return (a == b) | (np.isnan(a) & np.isnan(b))


def expand_window_to_grid(win, expand_pixels=(0, 0)):
    """
    Expand rasterio window extents to nearest whole numbers i.e. for expand_pixels >= (0, 0), it will return a window
    that contains the original extents.

    Parameters
    ----------
    win: rasterio.windows.Window
        The window to expand.
    expand_pixels: tuple, optional
        A tuple specifying the number of (rows, columns) pixels to expand the window by.
        [default: (0, 0)]

    Returns
    -------
    win: rasterio.windows.Window
         The expanded window.
    """
    col_off, col_frac = np.divmod(win.col_off - expand_pixels[1], 1)
    row_off, row_frac = np.divmod(win.row_off - expand_pixels[0], 1)
    width = np.ceil(win.width + 2 * expand_pixels[1] + col_frac)
    height = np.ceil(win.height + 2 * expand_pixels[0] + row_frac)
    exp_win = Window(col_off.astype('int'), row_off.astype('int'), width.astype('int'), height.astype('int'))
    return exp_win


def round_window_to_grid(win):
    """
    Round window extents to the nearest whole numbers.

    Parameters
    ----------
    win: rasterio.windows.Window
        The window to round.

    Returns
    -------
    win: rasterio.windows.Window
        The rounded window with integer extents.
    """
    row_range, col_range = win.toranges()
    row_range = np.round(row_range).astype('int')
    col_range = np.round(col_range).astype('int')
    return Window(col_off=col_range[0], row_off=row_range[0], width=np.diff(col_range)[0], height=np.diff(row_range)[0])


def validate_kernel_shape(kernel_shape, method=Method.gain_blk_offset):
    """
    Check a kernel_shape (height, width) tuple for validity.  Raises ValueError if kernel_shape is invalid.

    Parameters
    ----------
    kernel_shape: tuple
        The kernel (height, width) in pixels.
    method: Method, optional
        The modelling method kernel_shape will be used with.

    Returns
    -------
    kernel_shape: numpy.array
        The validated kernel_shape as a numpy array.
    """
    kernel_shape = np.array(kernel_shape).astype(int)
    if not np.all(np.mod(kernel_shape, 2) == 1):
        raise ValueError("kernel shape must be odd in both dimensions.")
    if method == Method.gain_offset and not np.product(kernel_shape) >= 25:
        raise ValueError("kernel shape area should contain at least 25 elements for the gain-offset method.")
    if not np.all(kernel_shape >= 1):
        raise ValueError("kernel shape must be a minimum of one in both dimensions.")
    return kernel_shape


def overlap_for_kernel(kernel_shape):
    """
    Return the block overlap for a kernel shape.

    Parameters
    ----------
    kernel_shape: tuple
        The kernel (height, width) in pixels.

    Returns
    -------
    overlap: numpy.array
        The overlap (height, width) in integer pixels as a numpy.array.
    """
    # Block overlap should be at least half the kernel 'shape' to ensure full kernel coverage at block edges, and a
    # minimum of (1, 1) to avoid including extrapolated (rather than interpolated) pixels when up-sampling.
    kernel_shape = np.array(kernel_shape).astype(int)
    return np.ceil(kernel_shape / 2).astype('int')


def validate_threads(threads):
    """Parse number of threads parameter."""
    _cpu_count = cpu_count()
    threads = _cpu_count if threads == 0 else threads
    if threads > _cpu_count:
        raise ValueError(f"'threads' is limited to the number of processors ({_cpu_count})")
    return threads


def create_homo_postfix(proc_crs, method, kernel_shape, driver='GTiff'):
    """Create a filename postfix, including extension, for the homogenised image file"""
    ext_dict = rio.drivers.raster_driver_extensions()
    ext_idx = list(ext_dict.values()).index(driver)
    ext = list(ext_dict.keys())[ext_idx]
    post_fix = f'_HOMO_c{proc_crs.name.upper()}_m{method.upper()}_k{kernel_shape[0]}_{kernel_shape[1]}.{ext}'
    return post_fix


def create_param_filename(filename: pathlib.Path):
    """Create a debug image filename, given the homogenised image filename"""
    filename = pathlib.Path(filename)
    return filename.parent.joinpath(f'{filename.stem}_PARAM{filename.suffix}')


def covers_bounds(im1, im2, expand_pixels=(0, 0)):
    """
    Determines if the spatial extents of one image cover another image

    Parameters
    ----------
    im1: rasterio.DatasetReader
        An open rasterio dataset.
    im2: rasterio.DatasetReader
        Another open rasterio dataset.
    expand_pixels: Tuple[int, int], optional
        Expand the im2 bounds by this many pixels.

    Returns
    -------
    covers_bounds: bool
        True if im1 covers im2 else False.
    """
    # use WarpedVRT to get the datasets in the same crs
    _im2 = WarpedVRT(im2, crs=im1.crs) if im1.crs != im2.crs else im2
    im1_win = im1.window(*_im2.bounds)
    if not np.all(np.array(expand_pixels) == 0):
        im1_win = expand_window_to_grid(im1_win, expand_pixels)
    win_ul = np.array((im1_win.row_off, im1_win.col_off))
    win_shape = np.array((im1_win.height, im1_win.width))
    return False if np.any(win_ul < 0) or np.any(win_shape > im1.shape) else True


def get_nonalpha_bands(im):
    """
    Return a list of non-alpha band indices from a rasterio dataset.

    Parameters
    ----------
    im: rasterio.DatasetReader
        Retrieve band indices from this dataset.

    Returns
    -------
    bands: list[int, ]
        The list of 1-based band indices.
    """
    bands = tuple([bi + 1 for bi in range(im.count) if im.colorinterp[bi] != ColorInterp.alpha])
    return bands


def combine_profiles(in_profile, config_profile):
    """
    Update an input rasterio profile with a configuration profile.

    Parameters
    ----------
    in_profile: dict
        The input/initial rasterio profile to update.  Driver-specific items are in the root dict.
    config_profile: dict
        The configuration profile.  Driver specific options are contained in a nested dict, with 'creation_options' key.
        E.g. see homonim.fuse.RasterFuse.default_out_profile.

    Returns
    -------
    out_profile: dict
        The combined profile.
    """

    if in_profile['driver'].lower() != config_profile['driver'].lower():
        # copy only non driver specific keys from input profile when the driver is different to the configured val
        copy_keys = ['driver', 'width', 'height', 'count', 'dtype', 'crs', 'transform']
        out_profile = {copy_key: in_profile[copy_key] for copy_key in copy_keys}
    else:
        out_profile = in_profile.copy()  # copy the whole input profile

    def nested_update(self_dict, other_dict):
        """Update self_dict with a flattened version of other_dict"""
        for other_key, other_value in other_dict.items():
            if isinstance(other_value, dict):
                # flatten the driver specific nested dict into the root dict
                nested_update(self_dict, other_value)
            elif other_value is not None:
                self_dict[other_key] = other_value
        return self_dict

    # update out_profile with a flattened config_profile
    return nested_update(out_profile, config_profile)



"""
    Homonim: Radiometric homogenisation of aerial and satellite imagery
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from pathlib import Path

from setuptools import setup, find_packages

"""
 Build and upload to testpypi:
     conda install -c conda-forge build twine
     python -m build
     python -m twine upload --repository testpypi dist/*

 Install from testpypi:
    python -m pip install --extra-index-url https://test.pypi.org/simple/ homonim

 Install local development version:
    pip install -e .
"""

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
version = {}
with open("homonim/version.py") as fp:
    exec(fp.read(), version)

setup(
    name="homonim",
    version=version['__version__'],
    description="Radiometric homogenisation of aerial and satellite imagery by fusion with satellite surface "
                "reflectance data.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Dugal Harris",
    author_email="dugalh@gmail.com",
    url="https://github.com/dugalh/homonim",
    license='AGPLv3',
    packages=find_packages(exclude=['tests', 'data'], include=['homonim']),
    install_requires=["numpy>=1.19", "rasterio>=1.1", "click>=8", "tqdm>=4.6", "pyyaml>=5.4",
                      "opencv-python-headless>=4.5", "pandas>=1.1"],
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": ["homonim=homonim.cli:cli"]},
)



"""
    Homonim: Radiometric homogenisation of aerial and satellite imagery
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Tuple

import cv2
import cv2 as cv
import numpy as np
from homonim import utils
from homonim.enums import Method
from homonim.raster_array import RasterArray
from rasterio.enums import Resampling
from rasterio.fill import fillnodata


class KernelModel:
    """
    A base class for surface reflectance modelling and homogenisation of blocks of image data.

    The surface reflectance relationship between source and reference image blocks is approximated with localised linear
    models.  Models are estimated for each pixel location inside a small rectangular kernel (window), using a fast DFT
    approach.  The homogenised output is produced by applying the model parameters to the source image.

    Based on the paper:
    Harris, Dugal & Van Niekerk, Adriaan. (2018). Radiometric homogenisation of aerial images by calibrating with
    satellite data. International Journal of Remote Sensing. 40. 1-25. 10.1080/01431161.2018.1528404.
    https://www.researchgate.net/publication/328317307_Radiometric_homogenisation_of_aerial_images_by_calibrating_with_satellite_data
    """

    default_config = dict(downsampling='average', upsampling='cubic_spline', r2_inpaint_thresh=0.25,
                          mask_partial=False)

    def __init__(self, method, kernel_shape, param_image=False, r2_inpaint_thresh=default_config['r2_inpaint_thresh'],
                 mask_partial=default_config['mask_partial'], downsampling=default_config['downsampling'],
                 upsampling=default_config['upsampling']):
        """
        Construct a KernelModel.

        Parameters
        ----------
        method: homonim.enums.Method
                The radiometric homogenisation method.
        kernel_shape: tuple
                The (height, width) of the kernel in pixels.
        param_image: bool, optional
                Turn on/off the calculation of Pearson's correlation coefficient (r2) for each kernel model.
                (Typically used for producing 'debug' images of the model parameters and accuracies.)
                [default: False]
        r2_inpaint_thresh: float, optional
                The R2 (coefficient of determination) threshold below which to 'in-paint' kernel model parameters from
                surrounding areas (applies to 'method' == Method.gain_offset only).  For pixels where the model gives a
                poor approximation to the data (this can occur in e.g. shadowed areas, poor source-reference
                co-registration, or areas where there has been land cover change between the source and reference
                images), model offsets are interpolated from surrounding areas, and gains re-estimated.  It can be set
                to 'None' to turn off in-painting.
                [default: 0.25]
        mask_partial: bool, optional
                Masks output pixels that were not produced by full kernel or source/reference image coverage.  Useful
                for ensuring strict model validity, and reducing seam-lines between adjacent images. [default: False]
        downsampling: rasterio.enums.Resampling
                The resampling method to use when downsampling.  [default: Resampling.average]
        upsampling: rasterio.enums.Resampling
                The resampling method to use when upsampling.  [default: Resampling.cubic_spline]
        """

        if not isinstance(method, Method):
            raise ValueError("'method' should be an instance of homonim.enums.Method")
        self._method = method

        self._kernel_shape = utils.validate_kernel_shape(kernel_shape, method=method)

        self._param_image = param_image
        self._r2_inpaint_thresh = r2_inpaint_thresh
        self._mask_partial = mask_partial
        self._downsampling = downsampling
        self._upsampling = upsampling

    @property
    def method(self) -> Method:
        """The homogenisation method."""
        return self._method

    @property
    def kernel_shape(self) -> Tuple[int, int]:
        """The kernel (height, width)."""
        return tuple(self._kernel_shape)

    @property
    def config(self) -> dict:
        """A dict containing the KernelModel configuration."""
        return dict(r2_inpaint_thresh=self._r2_inpaint_thresh, downsampling=self._downsampling,
                    upsampling=self._upsampling)

    def _r2_array(self, ref_array, src_array, param_array, mask=None, mask_sum=None, ref_sum=None, src_sum=None,
                  ref2_sum=None, src2_sum=None, src_ref_sum=None, dest_array=None, kernel_shape=None):
        """
        Helper function to find R2 coefficient of determination at each pixel/kernel location for the given matrices.
        """

        if kernel_shape is None:
            kernel_shape = self._kernel_shape
        kernel_shape = tuple(kernel_shape)  # force to tuple for opencv
        filter_args = dict(normalize=False, borderType=cv.BORDER_CONSTANT)  # common opencv arguments

        # find the keyword arguments that were not provided
        if mask is None:
            # if mask is passed, assume that it has been applied ref_array and src_array, otherwise do that here
            mask = (utils.nan_equals(src_array, RasterArray.default_nodata) &
                    utils.nan_equals(ref_array, RasterArray.default_nodata))
            ref_array[~mask] = 0
            src_array[~mask] = 0
        if mask_sum is None:
            mask_sum = cv.boxFilter(mask.astype(RasterArray.default_dtype), -1, kernel_shape, **filter_args)
        if ref_sum is None:
            ref_sum = cv.boxFilter(ref_array, -1, kernel_shape, **filter_args)
        if ref2_sum is None:
            ref2_sum = cv.sqrBoxFilter(ref_array, -1, kernel_shape, **filter_args)
        if src2_sum is None:
            src2_sum = cv.sqrBoxFilter(src_array, -1, kernel_shape, **filter_args)
        if src_ref_sum is None:
            src_ref_sum = cv.boxFilter(src_array * ref_array, -1, kernel_shape, **filter_args)

        # R2 is found using: R2 = 1 - (residual sum of squares)/(total sum of squares) = 1 - RSS/TSS

        # TSS = sum((ref - mean(ref))**2), which can be expanded and expressed in terms of cv2.boxFilter kernel sums as:
        ss_tot_array = (mask_sum * ref2_sum) - (ref_sum ** 2)

        if param_array.shape[0] > 1:
            # find RSS for method == Method.gain_offset
            if src_sum is None:
                src_sum = cv.boxFilter(src_array, -1, kernel_shape, **filter_args)

            # RSS = sum((ref - ref_hat)**2)
            #     = sum((ref - (m*src + c))**2), where m and c are the first 2 bands of param_array
            # The above can be expanded and expressed in terms of cv2.boxFilter kernel sums as:

            ss_res_array = (((param_array[0] ** 2) * src2_sum) +
                            (2 * np.product(param_array[:2], axis=0) * src_sum) -
                            (2 * param_array[0] * src_ref_sum) -
                            (2 * param_array[1] * ref_sum) +
                            ref2_sum + (mask_sum * (param_array[1] ** 2)))
        else:
            # find RSS for method == Method.gain or Method.gain_blk_offset
            # RSS = sum((ref - m*src)**2), where m is the first band of param_array
            # The above can be expanded and expressed in terms of cv2.boxFilter kernel sums as:
            ss_res_array = (((param_array[0] ** 2) * src2_sum) - (2 * param_array[0] * src_ref_sum) + ref2_sum)

        ss_res_array *= mask_sum

        if dest_array is None:
            # assign a destination array to write R2 into, if it was not provided
            dest_array = np.full(src_array.shape, fill_value=RasterArray.default_nodata,
                                 dtype=RasterArray.default_dtype)

        # find R2 = 1 - RSS/TSS, and write into dest_array
        np.divide(ss_res_array, ss_tot_array, out=dest_array, where=mask)
        np.subtract(1, dest_array, out=dest_array, where=mask)
        return dest_array

    @staticmethod
    def _fit_block_norm(ref_ra, src_ra):
        """
        Find a scale and offset for a source block, so that the standard deviation and first percentile of the source
        and reference blocks match.  (Can be thought of as a rough dark object subtraction (DOS) approach).

        Parameters
        ----------
        ref_ra : RasterArray
            Reference data block in a RasterArray.
        src_ra : RasterArray
            Source data block in a RasterArray, with the same CRS, shape & extents as ref_ra.

        Returns
        -------
        norm_model: numpy.array
            A two element [gain, offset] model to 'normalise' the source block.
        """
        norm_model = np.zeros(2)
        mask = ref_ra.mask & src_ra.mask
        if not np.any(mask):
            return norm_model
        norm_model[0] = np.std(ref_ra.array[mask]) / np.std(src_ra.array[mask])
        norm_model[1] = np.percentile(ref_ra.array[mask], 1) - np.percentile(src_ra.array[mask], 1) * norm_model[0]
        return norm_model

    def _fit_gain(self, ref_ra, src_ra, kernel_shape=None):
        """
        Find sliding kernel gains, for a band, using opencv convolution.

        Parameters
        ----------
        ref_ra : RasterArray
            Reference data block in a RasterArray.
        src_ra : RasterArray
            Source data block in a RasterArray, with the same CRS, shape & extents as ref_ra.
        kernel_shape : tuple, optional
            Sliding kernel [height, width] in pixels.

        Returns
        -------
        param_ra :RasterArray
            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
            R2 for each kernel model in the third band when param_image==True.
        """
        # adapted from https://www.mathsisfun.com/data/least-squares-regression.html with c=0
        if kernel_shape is None:
            kernel_shape = self._kernel_shape
        kernel_shape = tuple(kernel_shape)  # force to tuple for opencv

        # mask invalid pixels with 0 so that these do not contribute to kernel sums in *boxFilter()
        ref_array = ref_ra.array
        src_array = src_ra.array
        mask = ref_ra.mask & src_ra.mask
        ref_array[~mask] = 0
        src_array[~mask] = 0

        # setup a RasterArray profile for the parameters
        param_profile = src_ra.profile.copy()
        param_profile.update(count=3 if self._param_image else 2, nodata=RasterArray.default_nodata,
                             dtype=RasterArray.default_dtype)

        # convolve the kernel with src and ref to get kernel sums (uses DFT for large kernels)
        filter_args = dict(normalize=False, borderType=cv.BORDER_CONSTANT)  # common opencv arguments
        src_sum = cv.boxFilter(src_array, -1, kernel_shape, **filter_args)
        ref_sum = cv.boxFilter(ref_array, -1, kernel_shape, **filter_args)

        # create parameter RasterArray filled with nodata
        param_ra = RasterArray.from_profile(None, param_profile)
        param_ra.array[1, mask] = 0  # set offsets to 0

        # find sliding kernel gains, avoiding divide by 0
        np.divide(ref_sum, src_sum, out=param_ra.array[0], where=mask)

        if self._param_image:
            # Find R2 of the sliding kernel models
            self._r2_array(ref_array, src_array, param_ra.array[:1], mask=mask, ref_sum=ref_sum, src_sum=src_sum,
                           dest_array=param_ra.array[2], kernel_shape=kernel_shape)

        return param_ra

    def _fit_gain_blk_offset(self, ref_ra, src_ra, kernel_shape=None):
        """
        Find sliding kernel gains and 'image' (i.e. block) offset, for a band, using opencv convolution.

        ref_ra : RasterArray
            Reference data block in a RasterArray.
        src_ra : RasterArray
            Source data block in a RasterArray, with the same CRS, shape & extents as ref_ra.
        kernel_shape : tuple, optional
            Sliding kernel [height, width] in pixels.

        Returns
        -------
        param_ra :RasterArray
            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
            R2 for each kernel model in the third band when param_image==True.
        """
        if kernel_shape is None:
            kernel_shape = self._kernel_shape
        kernel_shape = tuple(kernel_shape)  # force to tuple for opencv

        # find the source->reference normalisation
        norm_model = self._fit_block_norm(ref_ra, src_ra)

        # force src nodata to nan so that operation below remains correctly masked
        src_ra.nodata = RasterArray.default_nodata

        # apply the normalisation (block gain and offset)
        src_ra.array = (src_ra.array * norm_model[0]) + norm_model[1]

        # find gains for normalised source
        param_ra = self._fit_gain(ref_ra, src_ra, kernel_shape=kernel_shape)

        # incorporate the normalisation model in the parameter RasterArray
        param_ra.array[1] = param_ra.array[0] * norm_model[1]
        param_ra.array[0] *= norm_model[0]
        return param_ra

    def _fit_gain_offset(self, ref_ra, src_ra, kernel_shape=None):
        """
        Find sliding kernel full linear model for a band using opencv convolution.

        ref_ra : RasterArray
            Reference data block in a RasterArray.
        src_ra : RasterArray
            Source data block in a RasterArray, with the same CRS, shape & extents as ref_ra.
        kernel_shape : tuple, optional
            Sliding kernel [height, width] in pixels.

        Returns
        -------
        param_ra :RasterArray
            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
            R2 for each kernel model in the third band when param_image==True.
        """
        # Least squares formulae adapted from https://www.mathsisfun.com/data/least-squares-regression.html
        if kernel_shape is None:
            kernel_shape = self._kernel_shape
        kernel_shape = tuple(kernel_shape)  # force to tuple for opencv

        # mask invalid pixels with 0 so that these do not contribute to kernel sums in *boxFilter()
        ref_array = ref_ra.array
        src_array = src_ra.array
        mask = ref_ra.mask & src_ra.mask
        ref_array[~mask] = 0
        src_array[~mask] = 0

        # setup a RasterArray profile for the parameters
        param_profile = src_ra.profile.copy()
        find_r2 = self._param_image or (self._r2_inpaint_thresh is not None)
        param_profile.update(count=3 if find_r2 else 2, nodata=RasterArray.default_nodata,
                             dtype=RasterArray.default_dtype)

        # find the numerator for the gain i.e N*cov(ref, src)
        filter_args = dict(normalize=False, borderType=cv.BORDER_CONSTANT)  # common opencv arguments
        src_sum = cv.boxFilter(src_array, -1, kernel_shape, **filter_args)
        ref_sum = cv.boxFilter(ref_array, -1, kernel_shape, **filter_args)
        src_ref_sum = cv.boxFilter(src_array * ref_array, -1, kernel_shape, **filter_args)
        mask_sum = cv.boxFilter(mask.astype(RasterArray.default_dtype, copy=False), -1, kernel_shape, **filter_args)
        m_num_array = (mask_sum * src_ref_sum) - (src_sum * ref_sum)

        # find the denominator for the gain i.e. N*var(src)
        src2_sum = cv.sqrBoxFilter(src_array, -1, kernel_shape, **filter_args)
        m_den_array = (mask_sum * src2_sum) - (src_sum ** 2)

        # create parameter RasterArray filled with nodata
        param_ra = RasterArray.from_profile(None, param_profile)

        # find the gain = cov(ref, src) / var(src), avoiding divide by 0
        np.divide(m_num_array, m_den_array, out=param_ra.array[0], where=mask)

        # solve for the offset c = y - mx, given that the linear model passes through (mean(ref), mean(src))
        np.divide(ref_sum - (param_ra.array[0] * src_sum), mask_sum, out=param_ra.array[1], where=mask)

        if find_r2:
            # Find R2 of the sliding kernel models
            self._r2_array(ref_array, src_array, param_ra.array[:2], mask=mask, mask_sum=mask_sum, ref_sum=ref_sum,
                           src_sum=src_sum, src2_sum=src2_sum, src_ref_sum=src_ref_sum, dest_array=param_ra.array[2],
                           kernel_shape=kernel_shape)

        if self._r2_inpaint_thresh is not None:
            # fill/inpaint low R2 and negative gain areas in the offset parameter
            r2_mask = (param_ra.array[2] > self._r2_inpaint_thresh) & (param_ra.array[0] > 0) & mask
            param_ra.array[1] = fillnodata(param_ra.array[1], r2_mask)
            param_ra.mask = mask  # re-mask as nodata areas will have been filled above

            # recalculate the gain for the filled areas using m = (y - c)/x and and the point (mean(ref), mean(src)))
            r2_mask = ~r2_mask & mask
            np.divide(ref_sum - mask_sum * param_ra.array[1], src_sum, out=param_ra.array[0], where=r2_mask)

        return param_ra

    def _full_coverage_mask(self, in_mask_ra, param_ra):
        """
        Helper function to create a full coverage mask i.e. a mask of output pixels fully covered by source/reference
        data, sliding (model) kernels, and resampling kernels.

        Note: that this is a strict approach that avoids any kind of partial coverage.

        Parameters
        ----------
        in_mask_ra: RasterArray
            An initial mask of valid source/reference pixels, as a RasterArray with dtype=='uint8'.
        param_ra: RasterArray
            A parameter RasterArray as returned by fit() i.e. a RasterArray in the CRS and grid corresponding to the
            'proc_crs' attribute, whose mask corresponds to the combined reference & source masks.
        Returns
        -------
        mask_ra: RasterArray
            The full coverage mask as a RasterArray with dtype=='uint8'
        """

        # re-project the initial mask into proc_crs (the CRS and grid corresponding to the proc_crs attribute)
        mask_ra = in_mask_ra.reproject(**param_ra.proj_profile, nodata=None, resampling=Resampling.average)
        # find the mask of proc_crs pixels fully covered by mask_ra
        mask = (mask_ra.array >= 1).astype('uint8', copy=False)  # ref pixels fully covered by src
        # combine mask with the other (src/ref) mask via param_ra
        mask &= param_ra.mask

        # Mask out partial kernel coverage.
        # Similar to the block overlap amount, this removes ceil(kernel_shape/2) pixels from the nodata edge.  Note,
        # that this is the strict approach for proc_crs == ref, it could be floor(kernel_shape/2) for proc_crs == src,
        # which avoids the additional upsampling step.
        se = cv2.getStructuringElement(cv2.MORPH_RECT, tuple(self._kernel_shape + 2))
        mask_ra.array = cv2.erode(mask, se, borderType=cv2.BORDER_CONSTANT, borderValue=0)
        return mask_ra

    def fit(self, ref_ra, src_ra, kernel_shape=None):
        """
        Fit sliding kernel models to reference and source blocks.

        Parameters
        ----------
        ref_ra : RasterArray
            Reference data block in a RasterArray.
        src_ra : RasterArray
            Source data block in a RasterArray, with the same CRS, shape & extents as ref_ra.
        kernel_shape : tuple, optional
            Sliding kernel [height, width] in pixels.

        Returns
        -------
        param_ra :RasterArray
            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
            R2 for each kernel model in the third band when param_image==True.
        """
        # TODO: include a CRS comparison below i.e. one that is faster that rasterio's current implementation
        if ((ref_ra.transform != src_ra.transform) or (ref_ra.shape != src_ra.shape)):
            raise ValueError("'ref_ra' and 'src_ra' must have the same transform and shape")

        if kernel_shape is None:
            kernel_shape = self._kernel_shape
        kernel_shape = tuple(kernel_shape)  # force to tuple for opencv

        if self._method == Method.gain:
            param_ra = self._fit_gain(ref_ra, src_ra, kernel_shape=kernel_shape)
        elif self._method == Method.gain_blk_offset:
            param_ra = self._fit_gain_blk_offset(ref_ra, src_ra, kernel_shape=kernel_shape)
        else:
            param_ra = self._fit_gain_offset(ref_ra, src_ra, kernel_shape=kernel_shape)

        return param_ra

    def apply(self, src_ra, param_ra):
        """
        Apply kernel models to a source block.

        Parameters
        ----------
        src_ra : RasterArray
            Source data block in a RasterArray.
        param_ra :RasterArray
            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and the same
            CRS, shape & extents as src_ra.

        Returns
        -------
        out_ra :RasterArray
            Homogenised block in a RasterArray.
        """
        out_array = (param_ra.array[0] * src_ra.array) + param_ra.array[1]
        out_ra = RasterArray.from_profile(out_array, param_ra.profile)
        return out_ra


class RefSpaceModel(KernelModel):
    """
    A KernelModel subclass, for estimating model parmaeters in the reference image CRS and grid.

    The source image block is re-projected into the reference CRS to estimate the parameters.  Estimated parameters are
    subsequently re-projected to the source CRS for application to the (original) source image block.

    Recommended for the most common use case where the reference image has a lower resolution than the source image.
    """

    def fit(self, ref_ra, src_ra, **kwargs):
        # choose resampling method based on whether we are up- or downsampling
        resampling = self._downsampling if np.prod(src_ra.res) < np.prod(ref_ra.res) else self._upsampling
        # downsample src_ra to reference CRS and grid
        src_ds_ra = src_ra.reproject(**ref_ra.proj_profile, resampling=resampling)
        # call base class fit with reference and source RasterArrays in the reference CRS & grid
        return KernelModel.fit(self, ref_ra, src_ds_ra, kernel_shape=self._kernel_shape)

    def apply(self, src_ra, param_ra):
        # create a parameter RasterArray containing only the first two bands of param_ra
        # (to speed up the re-projection below)
        _param_ra = RasterArray.from_profile(param_ra.array[:2], param_ra.profile)
        # choose resampling method based on whether we are up- or downsampling
        resampling = self._upsampling if np.prod(src_ra.res) < np.prod(param_ra.res) else self._downsampling
        # re-project param_ra to source CRS and grid
        param_us_ra = _param_ra.reproject(**src_ra.proj_profile, resampling=resampling)

        if self._mask_partial:
            # find the mask of fully covered pixels in reference CRS and grid
            mask_ra = self._full_coverage_mask(src_ra.mask_ra, _param_ra)
            # re-project the mask to source CRS and grid, and apply to the parameters
            mask_us_ra = mask_ra.reproject(**src_ra.proj_profile, nodata=0, resampling=Resampling.nearest)
            param_us_ra.mask = mask_us_ra.array.astype('bool', copy=False)
        else:
            param_us_ra.mask = src_ra.mask

        # call base class apply with source and parameter RasterArrays in the source CRS & grid
        return KernelModel.apply(self, src_ra, param_us_ra)


class SrcSpaceModel(KernelModel):
    """
    A KernelModel subclass, for estimating model parmaeters in the source image CRS and grid.

    The reference image block is re-projected into the source CRS to estimate the parameters.  Estimated parameters are
    subsequently applied (directly) to the source image block.

    Recommended for the unusual use case where the source image has a lower resolution than the reference image.
    """

    def fit(self, ref_ra, src_ra, **kwargs):
        # choose resampling method based on whether we are up- or downsampling
        resampling = self._upsampling if np.prod(src_ra.res) < np.prod(ref_ra.res) else self._downsampling
        # upsample ref_ra to the source CRS and grid
        ref_us_ra = ref_ra.reproject(**src_ra.proj_profile, resampling=resampling)

        _src_ra = src_ra.copy()  # copy the source to avoid in-place changes in fit() below
        # call base class fit with source and parameter RasterArrays in the source CRS & grid
        param_ra = KernelModel.fit(self, ref_us_ra, _src_ra, kernel_shape=self._kernel_shape)

        if self._mask_partial:
            # remove r2 band from param masking
            _param_ra = RasterArray.from_profile(param_ra.array[:2], param_ra.profile)
            # find the mask of fully covered pixels in source CRS and grid, and apply it to the parameters
            mask_ra = self._full_coverage_mask(ref_ra.mask_ra, _param_ra)
            param_ra.mask = mask_ra.array.astype('bool', copy=False)
        else:
            param_ra.mask = src_ra.mask

        return param_ra



"""
    Homonim: Radiometric homogenisation of aerial and satellite imagery
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import concurrent.futures
import logging
import pathlib
import threading
from collections import OrderedDict
from multiprocessing import cpu_count

import numpy as np
import rasterio as rio
import yaml
from rasterio.windows import get_data_window
from tqdm import tqdm

from homonim import utils
from homonim.enums import Method
from homonim.errors import ImageFormatError

logger = logging.getLogger(__name__)


class ParamStats:
    """Class to find statistics of a parameter image."""
    default_config = dict(threads=cpu_count() - 1)

    def __init__(self, param_filename, threads=default_config['threads']):
        """
        Construct the ParamStats object

        Parameters
        ----------
        param_filename: pathlib.Path, str
            Path to the parameter image file
        threads: int, optional
            The number of threads to use for concurrent processing of bands (requires more memory).  0 = use all cpus.
        """
        self._param_filename = pathlib.Path(param_filename)
        self._threads = utils.validate_threads(threads)

        self._validate_image(self._param_filename)

        # read some parameters from the metadata
        with rio.open(self._param_filename, 'r') as param_im:
            self._tags = param_im.tags()
            self._method = self._tags['HOMO_METHOD'].replace('_', '-')
            self._r2_inpaint_thresh = yaml.safe_load(self._tags['HOMO_MODEL_CONF'])['r2_inpaint_thresh']

    @staticmethod
    def _validate_image(param_filename):
        """Check file is a valid parameter image"""
        if not param_filename.exists():
            raise FileNotFoundError(f'{param_filename} does not exist')

        with rio.open(param_filename) as param_im:
            tags = param_im.tags()
            # check band count is a multiple of 3 and that expected metadata tags exist
            if (param_im.count == 0 or divmod(param_im.count, 3)[1] != 0 or
                    not {'HOMO_METHOD', 'HOMO_MODEL_CONF', 'HOMO_PROC_CRS'} <= set(tags)):
                raise ImageFormatError(f'{param_filename.name} is not a valid parameter image.')

            # check band descriptions end with the expected suffixes
            n_refl_bands = int(param_im.count / 3)
            suffixes = ['gain'] * n_refl_bands + ['offset'] * n_refl_bands + ['r2'] * n_refl_bands
            if not all([desc.lower().endswith(suffix) for suffix, desc in zip(suffixes, param_im.descriptions)]):
                raise ImageFormatError(f'{param_filename.name} is not a valid parameter image.')

    @property
    def metadata(self):
        """Return a printable string of parameter metadata"""
        res_str = (f'Method: {self._method}\n'
                   f'Kernel shape: {self._tags["HOMO_KERNEL_SHAPE"]}\n'
                   f'Processing CRS: {self._tags["HOMO_PROC_CRS"]}')
        if self._method == 'gain-offset':
            res_str += f'\nR\N{SUPERSCRIPT TWO} inpaint threshold: {self._r2_inpaint_thresh}'
        return res_str

    def stats(self):
        """
        Find parameter image statistics.

        Returns
        -------
        param_dict: dict[dict]
            A dictionary representing the results.
        """

        bar_format = '{l_bar}{bar}|{n_fmt}/{total_fmt} bands [{elapsed}<{remaining}]'
        read_lock = threading.Lock()

        with rio.Env(GDAL_NUM_THREADS='ALL_CPUs'), rio.open(self._param_filename) as param_im:
            # initialise the results dictionary in correct band order
            param_dict = OrderedDict([(param_desc, None) for param_desc in param_im.descriptions])

            # read mask once, assuming the same for all bands
            _mask = param_im.read_masks(indexes=1)

            # get window of valid pixels and re-read windowed mask
            valid_win = get_data_window(_mask, nodata=0)
            del _mask
            mask = param_im.read_masks(indexes=1, window=valid_win).astype('bool', copy=False)

            def process_band(band_i):
                with read_lock:
                    param_array = param_im.read(indexes=band_i + 1, window=valid_win, out_dtype='float32')
                param_vec = param_array[mask]  # vector of valid parameter values
                del param_array
                param_vec = np.ma.masked_invalid(param_vec).astype('float')  # mask out nan and inf values

                def stats(v: np.ma.core.MaskedArray):
                    """Find mean, std, min & max statistics for a vector."""
                    if v.mask.all():     # all values are invalid (nan / +-inf)
                        return OrderedDict(Mean=float('nan'), Std=float('nan'), Min=float('nan'), Max=float('nan'))
                    else:
                        return OrderedDict(Mean=v.mean(), Std=v.std(), Min=v.min(), Max=v.max())

                stats_dict = stats(param_vec)
                if (self._method == Method.gain_offset) and (band_i >= param_im.count * 2 / 3):
                    # Find the r2 inpaint portion if these parameters are from Method.gain_offset
                    inpaint_portion = np.sum(param_vec < self._r2_inpaint_thresh) / len(param_vec)
                    stats_dict['Inpaint (%)'] = inpaint_portion * 100

                return param_im.descriptions[band_i], stats_dict

            # loop over bands (i.e. parameters), finding statistics
            with concurrent.futures.ThreadPoolExecutor(max_workers=self._threads) as executor:
                futures = [executor.submit(process_band, band_i) for band_i in range(param_im.count)]

                # wait for threads, get results and raise any thread generated exceptions
                for future in tqdm(concurrent.futures.as_completed(futures), bar_format=bar_format, total=len(futures)):
                    param_desc, stats_dict = future.result()
                    param_dict[param_desc] = stats_dict

        return param_dict
